# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/chart.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.artifacts.v1 import chart_pb2 as chalk_dot_artifacts_dot_v1_dot_chart__pb2
from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2
from chalk._gen.chalk.chart.v1 import densetimeserieschart_pb2 as chalk_dot_chart_dot_v1_dot_densetimeserieschart__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1b\x63halk/server/v1/chart.proto\x12\x0f\x63halk.server.v1\x1a\x1e\x63halk/artifacts/v1/chart.proto\x1a\x1f\x63halk/auth/v1/permissions.proto\x1a)chalk/chart/v1/densetimeserieschart.proto\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto"L\n\x06Series\x12\x16\n\x06points\x18\x01 \x03(\x01R\x06points\x12\x14\n\x05label\x18\x02 \x01(\tR\x05label\x12\x14\n\x05units\x18\x03 \x01(\tR\x05units"t\n\x05\x43hart\x12\x14\n\x05title\x18\x01 \x01(\tR\x05title\x12/\n\x06series\x18\x02 \x03(\x0b\x32\x17.chalk.server.v1.SeriesR\x06series\x12$\n\x0ex_timestamp_ms\x18\x04 \x03(\x03R\x0cxTimestampMs"0\n\x05Point\x12\x19\n\x05value\x18\x01 \x01(\x03H\x00R\x05value\x88\x01\x01:\x02\x18\x01\x42\x08\n\x06_value"l\n\nTimeSeries\x12.\n\x06points\x18\x01 \x03(\x0b\x32\x16.chalk.server.v1.PointR\x06points\x12\x14\n\x05label\x18\x02 \x01(\tR\x05label\x12\x14\n\x05units\x18\x03 \x01(\tR\x05units:\x02\x18\x01"\xd7\x01\n\x0fTimeSeriesChart\x12\x14\n\x05title\x18\x01 \x01(\tR\x05title\x12\x33\n\x06series\x18\x02 \x03(\x0b\x32\x1b.chalk.server.v1.TimeSeriesR\x06series\x12\x35\n\x08x_series\x18\x03 \x03(\x0b\x32\x1a.google.protobuf.TimestampR\x07xSeries\x12>\n\rwindow_period\x18\x04 \x01(\x0b\x32\x19.google.protobuf.DurationR\x0cwindowPeriod:\x02\x18\x01"\xbe\x01\n\x11ListChartsFilters\x12P\n\x10link_entity_kind\x18\x01 \x01(\x0e\x32!.chalk.artifacts.v1.ChartLinkKindH\x00R\x0elinkEntityKind\x88\x01\x01\x12-\n\x10linked_entity_id\x18\x02 \x01(\tH\x01R\x0elinkedEntityId\x88\x01\x01\x42\x13\n\x11_link_entity_kindB\x13\n\x11_linked_entity_id"m\n\x12ListChartPageToken\x12@\n\x0e\x63reated_at_hwm\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0c\x63reatedAtHwm\x12\x15\n\x06id_hwm\x18\x02 \x01(\tR\x05idHwm"\xba\x01\n\x11ListChartsRequest\x12\x41\n\x07\x66ilters\x18\x01 \x01(\x0b\x32".chalk.server.v1.ListChartsFiltersH\x00R\x07\x66ilters\x88\x01\x01\x12\x19\n\x05limit\x18\x02 \x01(\x05H\x01R\x05limit\x88\x01\x01\x12"\n\npage_token\x18\x03 \x01(\tH\x02R\tpageToken\x88\x01\x01\x42\n\n\x08_filtersB\x08\n\x06_limitB\r\n\x0b_page_token"\xda\x01\n\x12ListChartsResponse\x12<\n\x06\x63harts\x18\x01 \x03(\x0b\x32 .chalk.artifacts.v1.MetricConfigB\x02\x18\x01R\x06\x63harts\x12\x45\n\x11\x63harts_with_links\x18\x02 \x03(\x0b\x32\x19.chalk.artifacts.v1.ChartR\x0f\x63hartsWithLinks\x12+\n\x0fnext_page_token\x18\x03 \x01(\tH\x00R\rnextPageToken\x88\x01\x01\x42\x12\n\x10_next_page_token"\xd2\x01\n\x17GetChartSnapshotRequest\x12\x45\n\rmetric_config\x18\x01 \x01(\x0b\x32 .chalk.artifacts.v1.MetricConfigR\x0cmetricConfig\x12\x39\n\nstart_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tstartTime\x12\x35\n\x08\x65nd_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x07\x65ndTime"\xcf\x01\n\x18GetChartSnapshotResponse\x12<\n\x06\x63harts\x18\x01 \x03(\x0b\x32$.chalk.chart.v1.DenseTimeSeriesChartR\x06\x63harts\x12\x35\n\x08x_series\x18\x02 \x03(\x0b\x32\x1a.google.protobuf.TimestampR\x07xSeries\x12>\n\rwindow_period\x18\x03 \x01(\x0b\x32\x19.google.protobuf.DurationR\x0cwindowPeriod2\xd9\x01\n\rChartsService\x12Z\n\nListCharts\x12".chalk.server.v1.ListChartsRequest\x1a#.chalk.server.v1.ListChartsResponse"\x03\x80}\x06\x12l\n\x10GetChartSnapshot\x12(.chalk.server.v1.GetChartSnapshotRequest\x1a).chalk.server.v1.GetChartSnapshotResponse"\x03\x80}\x06\x42\x93\x01\n\x13\x63om.chalk.server.v1B\nChartProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.chart_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\nChartProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_POINT"]._options = None
    _globals["_POINT"]._serialized_options = b"\030\001"
    _globals["_TIMESERIES"]._options = None
    _globals["_TIMESERIES"]._serialized_options = b"\030\001"
    _globals["_TIMESERIESCHART"]._options = None
    _globals["_TIMESERIESCHART"]._serialized_options = b"\030\001"
    _globals["_LISTCHARTSRESPONSE"].fields_by_name["charts"]._options = None
    _globals["_LISTCHARTSRESPONSE"].fields_by_name["charts"]._serialized_options = b"\030\001"
    _globals["_CHARTSSERVICE"].methods_by_name["ListCharts"]._options = None
    _globals["_CHARTSSERVICE"].methods_by_name["ListCharts"]._serialized_options = b"\200}\006"
    _globals["_CHARTSSERVICE"].methods_by_name["GetChartSnapshot"]._options = None
    _globals["_CHARTSSERVICE"].methods_by_name["GetChartSnapshot"]._serialized_options = b"\200}\006"
    _globals["_SERIES"]._serialized_start = 221
    _globals["_SERIES"]._serialized_end = 297
    _globals["_CHART"]._serialized_start = 299
    _globals["_CHART"]._serialized_end = 415
    _globals["_POINT"]._serialized_start = 417
    _globals["_POINT"]._serialized_end = 465
    _globals["_TIMESERIES"]._serialized_start = 467
    _globals["_TIMESERIES"]._serialized_end = 575
    _globals["_TIMESERIESCHART"]._serialized_start = 578
    _globals["_TIMESERIESCHART"]._serialized_end = 793
    _globals["_LISTCHARTSFILTERS"]._serialized_start = 796
    _globals["_LISTCHARTSFILTERS"]._serialized_end = 986
    _globals["_LISTCHARTPAGETOKEN"]._serialized_start = 988
    _globals["_LISTCHARTPAGETOKEN"]._serialized_end = 1097
    _globals["_LISTCHARTSREQUEST"]._serialized_start = 1100
    _globals["_LISTCHARTSREQUEST"]._serialized_end = 1286
    _globals["_LISTCHARTSRESPONSE"]._serialized_start = 1289
    _globals["_LISTCHARTSRESPONSE"]._serialized_end = 1507
    _globals["_GETCHARTSNAPSHOTREQUEST"]._serialized_start = 1510
    _globals["_GETCHARTSNAPSHOTREQUEST"]._serialized_end = 1720
    _globals["_GETCHARTSNAPSHOTRESPONSE"]._serialized_start = 1723
    _globals["_GETCHARTSNAPSHOTRESPONSE"]._serialized_end = 1930
    _globals["_CHARTSSERVICE"]._serialized_start = 1933
    _globals["_CHARTSSERVICE"]._serialized_end = 2150
# @@protoc_insertion_point(module_scope)



import argparse
import os
import pathlib
import typing
import warnings
from collections import defaultdict
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Type, Union, cast

from typing_extensions import Annotated, get_args, get_origin

from chalk.features import Feature, Features, FeatureSetBase, Vector, FeatureWrapper, unwrap_feature
from chalk.importer import import_all_python_files_from_dir
from chalk.utils.paths import get_directory_root

try:
    from types import UnionType
except ImportError:
    UnionType = None


class ParsedAnnotation(NamedTuple):
    annotation: str
    include_in_protocol_cls: bool
    protocol_annotation: str


class ParsedFeaturesClass(NamedTuple):
    module: str
    annotations: Dict[str, ParsedAnnotation]


class StubGenerator:
    def __init__(self) -> None:
        super().__init__()
        self._parsed_feature_classes: Dict[str, ParsedFeaturesClass] = {}
        self._module_to_stubs: List[Type[Features]] = []
        self._imports: Dict[Tuple[str, ...], str] = {}  # Mapping of full module name to the name it is imported as

    def _add_import(
        self,
        item: Union[str, type, typing._SpecialForm],  # pyright: ignore[reportPrivateUsage]
        import_as: Optional[str] = None,
    ) -> str:
        """Add an import for item.

        If `item` is already imported, then it will not be imported again.
        Its existing import name will be returned

        Parameters
        ----------
        item
            The thing to import. Should be the full path to the item (e.g. `typing.Optional`)

        Returns
        -------
        str
            The name that `item` will be imported as.
        """
        if isinstance(item, typing._SpecialForm):  # pyright: ignore[reportPrivateUsage]
            item = str(item)
        if isinstance(item, FeatureWrapper):
            item = unwrap_feature(item).fqn
        if isinstance(item, type):
            item = f"{item.__module__}.{item.__name__}"
        if not isinstance(item, str):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError(f"Invalid type for {item}: {type(item)}")
        item_split = tuple(item.split("."))
        if item_split[0] == "builtins":
            assert len(item_split) == 2
            return item_split[1]
        if item_split in self._imports:
            imported_name = self._imports[item_split]
            if import_as is not None and import_as != imported_name:
                raise ValueError(
                    f"Item {'.'.join(item_split)} already imported as {imported_name}; cannot be imported again as {import_as}"
                )
            return imported_name
        if import_as is None:
            import_as = self._get_import_name(item_split)
        self._imports[item_split] = import_as
        return self._imports[item_split]

    def _get_import_name(self, item_split: Tuple[str, ...]):
        if item_split[0] in ("typing", "typing_extensions"):
            assert len(item_split) == 2
            return item_split[1]
        item_split = tuple(x.replace("_", "__") for x in item_split)
        item = f"__stubgen_{'_'.join(item_split)}"
        return item

    def _register_imports_for_annotation(self, annotation: Any) -> Tuple[str, str]:
        """Register all necessary imports, and return a tuple of (protocol annotation, init annotation)"""
        if annotation == type(None):
            return "None", "None"
        if annotation is ...:
            return "...", "..."
        origin, args = get_origin(annotation), get_args(annotation)
        if origin is None:
            assert args == ()
            if isinstance(annotation, type) and issubclass(
                annotation, Vector
            ):  # pyright: ignore[reportUnnecessaryIsInstance]
                origin_name = self._add_import(Vector, import_as="Vector")
                return "Vector", "Vector"
            elif annotation.__module__ == "builtins":
                assert isinstance(annotation, type)
                return annotation.__name__, annotation.__name__
            else:
                ans = self._add_import(annotation)
                return ans, ans
        # Some sort of generic class, like a collection, or optional
        if origin == UnionType:
            # Convert new-style type unions to old-style typing.Unions
            origin = typing.Union
        if origin == Annotated:
            # If annotated, only pay attention to the first arg -- the actual type
            if args[1] == "__chalk_primary__":
                protocol_arg_name, init_arg_name = self._register_imports_for_annotation(args[0])
                return f"Primary[{protocol_arg_name}]", init_arg_name
            return self._register_imports_for_annotation(args[0])
        origin_name = self._add_import(cast(Any, origin))
        arg_names = [self._register_imports_for_annotation(arg) for arg in args]
        return (
            f"{origin_name}[" + ", ".join(x[0] for x in arg_names) + "]",
            f"{origin_name}[" + ", ".join(x[1] for x in arg_names) + "]",
        )

    def register_features_class(self, features_cls: Type[Features]):
        annotations: Dict[str, ParsedAnnotation] = {}
        for f in features_cls.features:
            assert isinstance(f, Feature)
            assert f.typ is not None
            if f.typ.as_dataframe() is not None:
                self._add_import("chalk.features.DataFrame", "DataFrame")
                annotations[f.attribute_name] = ParsedAnnotation(
                    "DataFrame",
                    include_in_protocol_cls=True,
                    protocol_annotation="DataFrame",
                )
            elif isinstance(f.typ._underlying, Feature):  # pyright: ignore[reportPrivateUsage]
                continue
            else:
                # Make a best-effort to figure out imports.
                proto_annotation, init_annotation = self._register_imports_for_annotation(f.typ.parsed_annotation)
                include_in_proto_cls = not f.is_windowed_pseudofeature and not f.is_autogenerated
                if include_in_proto_cls:
                    protocol_annotation = proto_annotation
                    if f.typ.as_features_cls() is not None:
                        protocol_annotation = "Any"
                    else:
                        as_feature = f.typ.as_feature()
                        if as_feature is not None:
                            protocol_annotation, init_annotation = self._register_imports_for_annotation(
                                as_feature.typ.parsed_annotation
                            )
                    annotations[f.attribute_name] = ParsedAnnotation(
                        init_annotation,
                        include_in_protocol_cls=include_in_proto_cls,
                        # If it's a features class (i.e. a has-one), not specifying
                        # the type in the protocol class so the type checker can
                        # still match feature classes that have a circular reference
                        protocol_annotation="Any" if (f.typ.as_features_cls() is not None) else protocol_annotation,
                    )
        name = features_cls.__chalk_namespace__.replace("::", "_")
        name_parts = name.split("_")
        name = "".join(x.title() for x in name_parts)
        if name in self._parsed_feature_classes:
            # We could have duplicates because of namespacing. That is fine because we do protocol matching so the actual name doesn't matter
            raise ValueError(
                f"Unable to generate stubs due to multiple features sets named '{features_cls.__chalk_namespace__}'. Feature set names must be unique. Please rename one of these feature sets."
            )
        self._parsed_feature_classes[name] = ParsedFeaturesClass(
            module=features_cls.__module__,
            annotations=annotations,
        )

    def generate_features_decorator_stub_file(self):
        self._add_import("typing.Optional", "Optional")
        self._add_import("typing.Any", "Any")
        self._add_import("typing.Protocol", "Protocol")
        self._add_import("typing.Type", "Type")
        self._add_import("typing.overload", "overload")
        self._add_import("typing.Iterator", "Iterator")
        self._add_import("chalk.features.feature_set.FeaturesMeta", "FeaturesMeta")
        self._add_import("chalk.features.Features", "Features")
        self._add_import("chalk.features.TPrimitive", "TPrimitive")
        self._add_import("chalk.features.Feature", "Feature")
        self._add_import("chalk.features.primary.Primary", "Primary")
        self._add_import("chalk.utils.duration.Duration", "Duration")
        self._add_import("chalk.features.Tags", "Tags")
        lines = [
            "# AUTO-GENERATED FILE. Do not edit. Run `chalk stubgen` to generate.",
            "# fmt: off",
            "# isort: skip_file",
            "from __future__ import annotations",
            "",
        ]
        # sort the imports for consistency
        imports = sorted(self._imports.items())
        for imp, import_as in imports:
            module = imp[:-1]
            name = imp[-1]
            lines.append(f"from {'.'.join(module)} import {name} as {import_as}")
        lines.append("")

        # The overall strategy is to create 3 stub classes for each features class.
        #
        # In no particular order:
        #
        # Class 1: The metaclass. The metaclass includes @property descriptors for each class variable.
        # These descriptors are used to annotation class attributes as Type[feature]
        # Pyright requires that all annotations are types, not instances.
        #
        # Class 2: The class. The class includes an __init__ signature that includes all features, with the
        # correct type annotation. All features have a default value of `...`, since Chalk does not require feature
        # sets to be complete. Inside the __init__ body, all instance attributes are annotated. Placing the annotation
        # here, as opposed to the class body, does not conflict with the metaclass annotations
        #
        # Class 3: The protocol class. Since the @features decorator is passed the end user's class definition, and returns the
        # Chalk-ified version of it, we cannot "import" the user's class definition. So, we define a protocol class that contains the
        # same instance attributes. The type checker can match the user's class against the protocol.

        # Since we effectively duck type with Protocols, different feature sets with all the same attributes
        # and types cannot be distinguished
        # So, we will keep just one feature set for a given set of attributes and annotations
        attributes_and_annotations_to_feature_set_names: dict[frozenset[tuple[str, str]], list[str]] = defaultdict(list)
        for x_name, x in self._parsed_feature_classes.items():
            attributes_and_annotations_to_feature_set_names[
                frozenset((k, v.protocol_annotation) for (k, v) in x.annotations.items())
            ].append(x_name)
        feature_set_names_to_keep: set[str] = set()
        for feature_set_names in attributes_and_annotations_to_feature_set_names.values():
            # Sorting for stability
            feature_set_names_to_keep.add(sorted(feature_set_names)[0])
        self._parsed_feature_classes = {
            k: v for (k, v) in self._parsed_feature_classes.items() if k in feature_set_names_to_keep
        }

        # Update the base classes
        # Feature sets that are supersets of other feature classes must inherit from the base classes, due to how
        # @overload requires return and parameter types of the less specific classes to be supertypes
        parsed_features_class_name_to_bases: dict[str, list[str]] = {}
        for x_name, x in self._parsed_feature_classes.items():
            common_base_class_names: set[str] = set()
            for y_name, y in self._parsed_feature_classes.items():
                if x is not y:
                    if all(
                        x.annotations.get(attribute_name) == annotation
                        for attribute_name, annotation in y.annotations.items()
                    ):
                        # If all of y's annotations are a subset of x's, it's a possible base class
                        common_base_class_names.add(y_name)
            # Sorting the base classes in reverse order so the most specific will go first
            common_base_classes_sorted = sorted(
                common_base_class_names,
                key=lambda y_name: (len(self._parsed_feature_classes[y_name].annotations), y_name),
                reverse=True,
            )
            parsed_features_class_name_to_bases[x_name] = [*common_base_classes_sorted, "Features"]

        # An @overload is used to match each protocol class (class #3) to the typestub version of the class (class #2)
        # First, define all metaclasses, then the classes, then the protocols, most general to most specific
        # However, when defining the @overloads, go from most specific to most general, since the first matching @overload
        # is used
        # Also including the class name as part of the sort key for stability
        parsed_feature_classes = sorted(
            self._parsed_feature_classes.items(),
            key=lambda x: (len(x[1].annotations), x[0]),
        )

        for feature_cls_name, features_cls in parsed_feature_classes:
            # First generate the metaclass definition
            metaclasses_bases = ", ".join(
                (*(f"{base_class}Meta" for base_class in parsed_features_class_name_to_bases[feature_cls_name]), "type")
            )
            lines.append(
                f"""\
class {feature_cls_name}Meta({metaclasses_bases}):"""
            )
            for attribute, annotation in features_cls.annotations.items():
                lines.append(
                    f"""\
    @property
    def {attribute}(self) -> Type[{annotation.annotation}]: ...
"""
                )
            if len(features_cls.annotations) == 0:
                lines.append(
                    f"""\
    ...
"""
                )
            lines.append(
                f"""\
    def __iter__(self) -> Iterator[Feature[TPrimitive,Any]]: ...
"""
            )

        for feature_cls_name, features_cls in parsed_feature_classes:
            # Next, generate the class definition
            bases = ", ".join(parsed_features_class_name_to_bases[feature_cls_name])
            lines.append(
                f"""\
class {feature_cls_name}({bases}, metaclass={feature_cls_name}Meta):"""
            )
            lines.append(
                f"""\
    def __init__(
        self,"""
            )
            for attribute, annotation in features_cls.annotations.items():
                lines.append(
                    f"""\
        {attribute}: {annotation.annotation} = ...,"""
                )
            lines.append(
                """\
    ):"""
            )
            for attribute, annotation in features_cls.annotations.items():
                lines.append(
                    f"""\
        self.{attribute}: {annotation.annotation}"""
                )
            if len(features_cls.annotations) == 0:
                lines.append(
                    f"""\
        ..."""
                )
            lines.append("")
        for feature_cls_name, features_cls in parsed_feature_classes:
            # Finally, generate the protocol class representing the features class that is being transformed
            lines.append(
                f"""\
class {feature_cls_name}Protocol(Protocol):"""
            )
            has_proto_annotation = False
            for attribute, annotation in features_cls.annotations.items():
                if not annotation.include_in_protocol_cls:
                    continue
                has_proto_annotation = True
                lines.append(
                    f"""\
    {attribute}: {annotation.protocol_annotation}"""
                )
            if not has_proto_annotation:
                lines.append(
                    f"""\
    ..."""
                )
            lines.append("")

        # Add overloads to the @features decorator to hint that it returns
        # instances of the classes defined above, not the classes defined in the user's module
        # This allows us to put all type stubs in one file, under the chalk/features/feature namespace,
        # rather than having to create a type stub for each user's feature class.
        proto_sorted_feature_classes = sorted(
            self._parsed_feature_classes.items(),
            key=lambda x: (
                sum(1 for annotation in x[1].annotations.values() if annotation.include_in_protocol_cls),
                x[0],
            ),
            reverse=True,
        )
        for features_cls_name, features_cls in proto_sorted_feature_classes:
            # For each features class, adding two overloads. The first overload transforms the definition
            # from the user's code into our stub. However, since we are technically importing the version that
            # has already been transformed with @features, we need to annotate how to handle a class that
            # has already been processed -- hence the second definition which is effectively a no-op.
            lines.append(
                f"""\
@overload
def features(item: Type[{features_cls_name}Protocol]) -> Type[{features_cls_name}]: ...
"""
            )

        # We also need to handle when the user annotates the features class with args, such as an owner or tags, e.g.
        # @features(owner=...)
        # class MyFeaturesClass:
        #     ...
        # To do this, we add one additional overload for features that returns a protocol class.
        # This protocol class implements __call__ with overloads, similar to the above
        lines.append(
            """\
@overload
def features(
    *,
    owner: Optional[str] = None,
    tags: Optional[Tags] = None,
    etl_offline_to_online: bool = ...,
    max_staleness: Optional[Duration] = ...,
    name: Optional[str] = None,
) -> __stubgen__features_proto: ...

class __stubgen__features_proto(Protocol):"""
        )
        for features_cls_name, features_cls in proto_sorted_feature_classes:
            overload = (
                "@overload" if len(parsed_feature_classes) > 1 else ""
            )  # It is incorrect to use @overload if there is just one definition
            lines.append(
                f"""\
    {overload}
    def __call__(self, item: Type[{features_cls_name}Protocol]) -> Type[{features_cls_name}]: ...
"""
            )
        if len(parsed_feature_classes) == 0:
            lines.append(
                """\
    ...
"""
            )
        return lines


def configure_stubgen_argparse(parser: argparse.ArgumentParser):
    parser.add_argument("--file_filter", help="Path containing only files to consider", nargs="?")
    parser.add_argument(
        "--root",
        type=str,
        default=None,
        help="Project root to scan for features definitions. By default, will find the directory root.",
    )
    parser.add_argument(
        "--stub_path",
        "-s",
        type=str,
        default=None,
        help="Folder containing custom type stubs. By default, will use the `typings` folder, relative to the project root",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        default=False,
        help="Whether to print verbose error messages when stubgen fail",
    )


def run_stubgen(
    args: argparse.Namespace,
    file_filter: Optional[str],
):
    root = args.root
    if root is None:
        root = get_directory_root() or pathlib.Path(os.getcwd())
    else:
        assert isinstance(root, str)
        root = os.path.abspath(root)
        root = pathlib.Path(root)
    stub_path = args.stub_path
    if stub_path is None:
        stub_path = str(root / "typings")
    assert isinstance(stub_path, str)
    file_allowlist = None
    if file_filter is not None and file_filter != "":
        with open(file_filter) as file_filter_contents:
            file_allowlist = [
                pathlib.Path(x.strip()) for x in file_filter_contents.readlines() if x.strip().endswith(".py")
            ]
    failed_imports = import_all_python_files_from_dir(
        project_root=root,
        file_allowlist=file_allowlist,
    )
    if len(failed_imports) > 0:
        warnings.warn(
            f"Stubs may be incomplete due to errors in loading the following files: {', '.join([x.filename for x in failed_imports])}"
        )
        if args.verbose:
            for x in failed_imports:
                print(f"Failed import: {x.filename}: {x.traceback}")
    stubgen = StubGenerator()
    for features_cls in FeatureSetBase.registry.values():
        try:
            stubgen.register_features_class(features_cls)
        except Exception as e:
            raise RuntimeError(f"Error processing features class {features_cls}") from e
    lines = stubgen.generate_features_decorator_stub_file()
    folder = os.path.join(stub_path, "chalk", "features")
    os.makedirs(folder, exist_ok=True)
    output_file = os.path.join(folder, "feature_set_decorator.pyi")
    with open(output_file, "w+") as f:
        f.write("\n".join(lines))
    print(f"Successfully wrote type stubs to {output_file}")



from __future__ import annotations

from typing import Any, List, Literal, Tuple, Type, Union, overload

import numpy as np
import pyarrow as pa

from chalk.features.filter import Filter


class Vector:
    """The Vector class can be used type annotation to denote a Vector feature.

    Instances of this class will be provided when working with raw vectors inside of resolvers.
    Generally, you do not need to construct instances of this class directly, as Chalk will
    automatically convert list-like features into `Vector` instances when working with a `Vector`
    annotation.

    Parameters
    ----------
    data: numpy.Array | list[float] | pyarrow.FixedSizeListScalar
        The vector values

    Examples
    --------
    >>> from chalk.features import Vector, features
    >>> @features
    ... class Document:
    ...     embedding: Vector[1536]
    """

    precision: Literal["fp16", "fp32", "fp64"]
    """The precision of the Vector"""

    num_dimensions: int
    """The number of dimensions in the vector"""

    dtype: pa.DataType
    """The PyArrow data type for the Vector feature. It will be an pa.FixedSizeListType()"""

    def __init__(self, data: Union[pa.FixedSizeListScalar, np.ndarray]):
        super().__init__()
        if not isinstance(data, pa.FixedSizeListScalar):
            if isinstance(data, np.ndarray):
                if data.dtype.name == "float16":
                    dtype = pa.float16()
                elif data.dtype.name == "float32":
                    dtype = pa.float32()
                elif data.dtype.name == "float64":
                    dtype = pa.float64()
                else:
                    raise TypeError(
                        f"Unsupported type '{data.dtype.name}'. Vector features must be float16, float32, or float64."
                    )
                data_converted = pa.scalar(data, pa.list_(dtype, len(data)))
                assert isinstance(data_converted, pa.FixedSizeListScalar)
                data = data_converted
            else:
                raise TypeError(
                    f"When constructing a vector, the value must be a pa.FixedSizeListScalar or a numpy array. Got {data}"
                )
        self._data: pa.FixedSizeListScalar = data
        self.dtype = self._data.type
        value_type = self._data.type.value_type
        self.num_dimensions = self._data.type.list_size
        if value_type == pa.float16():
            self.precision = "fp16"
        elif value_type == pa.float32():
            self.precision = "fp32"
        elif value_type == pa.float64():
            self.precision = "fp64"
        else:
            raise TypeError(f"Value type {value_type} is unsupported. It must be a float16, float32, or float64.")

    def to_arrow_scalar(self) -> pa.FixedSizeListScalar:
        """Convert a vector to a PyArrow array.

        Returns
        -------
        pa.FixedSizeListArray
            The vector, as a PyArrow array.
        """
        return self._data

    def to_arrow_array(self) -> pa.Array:
        """Convert a vector to a PyArrow array.

        Returns
        -------
        pa.FixedSizeListArray
            The vector, as a PyArrow array.
        """
        return self._data.values

    def to_numpy(self, writable: bool = False) -> np.ndarray:
        """Convert the vector to a Numpy array.

        Parameters
        ----------
        writable: bool
            Whether the numpy array should be writable. If so, an extra copy of the vector data will be made.

        Returns
        -------
        np.ndarray
            The vector, as a numpy array.
        """
        return self._data.values.to_numpy(writable=writable)

    def to_pylist(self) -> List[float]:
        """Convert the vector to a Python list.

        Returns
        -------
        List[float]
            The vector, as a list of Python floats
        """
        return self._data.as_py()

    @overload
    def __class_getitem__(cls, item: Tuple[Literal["fp32", "fp64", "fp16"], int], /) -> Type[Vector]:
        """Create a Vector type with the specified precision and number of dimensions.

        Parameters
        ----------
        item: Tuple[Literal["fp32", "fp64", "fp16"], int]
            The first element is the precision, and the second element is the number of dimensions.

        Returns
        -------
        Type[Vector]
            A vector type
        """
        ...

    @overload
    def __class_getitem__(cls, item: int, /) -> Type[Vector]:
        """Create a Vector type of with float-16 precision and the specified number of dimensions.
        This helper is equivalent to `Vector['fp16', N]`.

        Parameters
        ----------
        item: int
            The number of dimensions.

        Returns
        -------
        Type[Vector]
            A vector type
        """
        ...

    def __class_getitem__(cls, item: Union[Tuple[Literal["fp32", "fp64", "fp16"], int], int], /) -> Type[Vector]:
        if isinstance(item, int):
            dimensions = item
            _precision = "fp16"
        elif isinstance(item, tuple) and len(item) == 2:  # pyright: ignore[reportUnnecessaryIsInstance]
            _precision, dimensions = item
        else:
            raise ValueError(
                f"Expected the Vector type argument to be the dimensions (e.g. Vector[1536]) or the dtype and dimensions (e.g. Vector['fp32', 1536]); got {item}"
            )

        if _precision == "fp16":
            _dtype = pa.list_(pa.float16(), dimensions)
        elif _precision == "fp32":
            _dtype = pa.list_(pa.float32(), dimensions)
        elif _precision == "fp64":
            _dtype = pa.list_(pa.float64(), dimensions)
        else:
            raise ValueError(f"Expected the dtype to be one of 'fp16', 'fp32', or 'fp64'; got {_precision}.")
        if not isinstance(dimensions, int):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError(f"Expected the dimensions to be an integer; got {dimensions}.")

        class SubclassedVector(Vector):
            precision = _precision
            num_dimensions = dimensions
            dtype = _dtype

            def __new__(cls, *args: Any, **kwargs: Any):
                return Vector(*args, **kwargs)

        return SubclassedVector

    @classmethod
    def is_near(
        cls,
        other: Any,
        metric: Literal["l2", "ip", "cos"] = "l2",
    ) -> Filter:
        """Define a nearest neighbor relationship for performing Vector similarity search.

        Parameters
        ----------
        other: Type[Vector]
            The other vector feature. This vector must have the same dtype and dimensions.

        metric: "l2" | "ip" | "cos"
            The metric to use to compute distance between two vectors. L2 Norm ("l2"), Inner Product ("ip"),
            and Cosine ("cos") are supported. Defaults to "l2".

        Returns
        -------
        Filter:
            A nearest neighbor relationship filter.
        """
        # This method is never called, since is_near is always called on the FeatureWrapper
        # Defining it here for type checking support
        raise NotImplementedError



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class FeaturePermission(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    FEATURE_PERMISSION_UNSPECIFIED: _ClassVar[FeaturePermission]
    FEATURE_PERMISSION_ALLOW: _ClassVar[FeaturePermission]
    FEATURE_PERMISSION_ALLOW_INTERNAL: _ClassVar[FeaturePermission]
    FEATURE_PERMISSION_DENY: _ClassVar[FeaturePermission]

FEATURE_PERMISSION_UNSPECIFIED: FeaturePermission
FEATURE_PERMISSION_ALLOW: FeaturePermission
FEATURE_PERMISSION_ALLOW_INTERNAL: FeaturePermission
FEATURE_PERMISSION_DENY: FeaturePermission

class FeaturePermissions(_message.Message):
    __slots__ = ("tags",)
    class TagsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: FeaturePermission
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[_Union[FeaturePermission, str]] = ...
        ) -> None: ...

    TAGS_FIELD_NUMBER: _ClassVar[int]
    tags: _containers.ScalarMap[str, FeaturePermission]
    def __init__(self, tags: _Optional[_Mapping[str, FeaturePermission]] = ...) -> None: ...



from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from chalk.client import DatasetRevision


class BaseCatalog(Protocol):
    def write_to_catalog(self, revision: "DatasetRevision", destination: str) -> None:
        ...



from typing import Any

from rich import reconfigure
from rich.control import Control
from rich.live import Live

from chalk._reporting.rich.theme import CHALK_THEME
from chalk.utils.notebook import is_notebook


class ChalkLive(Live):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        reconfigure(
            force_jupyter=False,
            force_terminal=True,
            color_system="truecolor",
            theme=CHALK_THEME,
            # `self.console.width` is a width that is
            # wider than if `width` is not specified
            # at all. We want the wider width so that
            # long resolver names are not truncated.
            width=int(self.console.width),
        )

    def refresh(self) -> None:
        """
        Chalk's custom implementation of refresh, to achieve
        a dark background for certain IDEs.
        """
        with self._lock:
            self._live_render.set_renderable(self.renderable)
            if self.console.is_jupyter:  # pragma: no cover
                try:

                    from IPython.display import clear_output, display
                    from ipywidgets import Output
                except ImportError:
                    import warnings

                    warnings.warn('install "ipywidgets" for Jupyter support')
                else:
                    if self.ipy_widget is None:
                        self.ipy_widget = Output()
                        display(self.ipy_widget)
                    assert self.ipy_widget is not None
                    with self.ipy_widget:
                        self.ipy_widget.clear_output(wait=True)
                        self.console.print(self._live_render.renderable)
            elif self.console.is_terminal and not self.console.is_dumb_terminal:
                if is_notebook():
                    from IPython.display import clear_output

                    clear_output(wait=True)
                with self.console:
                    self.console.print(Control())
            elif (
                not self._started and not self.transient
            ):  # if it is finished allow files or dumb-terminals to see final result
                with self.console:
                    self.console.print(Control())



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Union, cast

import isodate

from chalk.features._encoding.missing_value import MissingValueStrategy
from chalk.features.feature_field import Feature, FeatureNotFoundException
from chalk.utils.collections import get_unique_item

if TYPE_CHECKING:
    import polars as pl


def _polars_dtype_contains_struct(dtype: pl.PolarsDataType):
    """Returns whether the dtype contains a (potentially nested) struct"""
    import polars as pl

    if isinstance(dtype, pl.Struct) or (isinstance(dtype, type) and issubclass(dtype, pl.Struct)):
        return True
    if isinstance(dtype, pl.List):
        assert dtype.inner is not None
        return _polars_dtype_contains_struct(dtype.inner)
    return False


def _generate_empty_series_for_dtype(name: str, dtype: pl.PolarsDataType, length: int) -> pl.Series:
    """Safely generate a series of all null values for the specified datatype.

    Unlike the ``pl.Series`` constructor, this function can handle struct dtypes.
    """
    import polars as pl

    # if isinstance(dtype, pl.Struct):
    #     # Struct dtypes cannot be specified in the pl.Series constructor.
    #     # Instead, create a dataframe, then call .to_struct() on it
    #     # If recursing within a struct, it should have a length of zero
    #     data = {f.name: _generate_empty_series_for_dtype(f.name, f.dtype, length) for f in dtype.fields}
    #     temp_df = pl.DataFrame(data)
    #     return temp_df.to_struct(name)
    # if isinstance(dtype, pl.List) and isinstance(dtype.inner, pl.Struct):
    #     # The pl.Series constructor does not respect nested data types
    #     # So, we'll manually build a dataframe with the list set to the correct type,
    #     # set it to None, and then select just the list
    #     assert dtype.inner is not None
    #     data = {name: _generate_empty_series_for_dtype(name, dtype.inner, 0)}
    #     temp_df = pl.DataFrame(data)
    #     df_with_list = temp_df.select(pl.col(name).reshape((length, -1)))
    #     df_with_list = df_with_list.select(pl.when(True).then(None).otherwise(pl.col(name)).alias(name))
    #     list_of_struct_series = df_with_list.select(pl.col(name).reshape((length, -1))).get_column(name)
    #     return list_of_struct_series
    return pl.Series(name, dtype=dtype, values=([None] * length))


def _get_expected_dtype(ft: Feature) -> pl.PolarsDataType:
    import polars as pl

    dtype = ft.converter.polars_dtype
    for path_obj in ft.path:
        if path_obj.parent.is_has_many:
            dtype = pl.List(inner=dtype)
    return dtype


def validate_df_schema(underlying: Union[pl.DataFrame, pl.LazyFrame]):
    # This is called from within DataFrame.__init__, which validates that polars is installed
    import polars as pl

    for root_fqn, actual_dtype in underlying.schema.items():
        feature = Feature.from_root_fqn(root_fqn)
        if feature.is_has_one or feature.is_has_many:
            continue
        expected_dtype = _get_expected_dtype(feature)
        if actual_dtype == expected_dtype:
            continue
        if isinstance(underlying, pl.LazyFrame):
            underlying = underlying.collect()
        if len(underlying) == underlying.get_column(root_fqn).null_count() and actual_dtype != expected_dtype:
            # If all values are null, then replace with a null series of the correct datatype
            # It's quite possible that the original column will have an incorrect dtype if all values are null,
            # since there was no data to infer the correct dtype from
            underlying = underlying.with_columns(
                _generate_empty_series_for_dtype(root_fqn, expected_dtype, len(underlying))
            )
        elif (
            isinstance(expected_dtype, pl.List)
            and actual_dtype == pl.Utf8  # pyright: ignore[reportUnnecessaryComparison]
        ):
            col = pl.col(root_fqn).str.json_extract(expected_dtype)
            try:
                underlying = underlying.with_columns(col.cast(expected_dtype))
            except (Exception, pl.PolarsPanicError) as e:
                raise TypeError(
                    f"Values for list feature `{root_fqn}` could not be deserialized from a JSON string to dtype `{expected_dtype}`."
                ) from e

            deser_actual_dtype = underlying[root_fqn].dtype
            if deser_actual_dtype != expected_dtype:
                raise TypeError(
                    f"Values for list feature `{root_fqn}` could not be deserialized from a JSON string to dtype `{expected_dtype}`. Found type {deser_actual_dtype}, instead."
                )
        elif _polars_dtype_contains_struct(expected_dtype):
            # Cannot cast to a struct type. Instead, will error, so the user can ensure the underlying
            # dictionaries / dataclasses are of the correct type
            col = pl.col(root_fqn).cast(expected_dtype)
            try:
                underlying = underlying.with_columns(underlying.select(col))
            except (pl.ComputeError, pl.PolarsPanicError) as e:
                try:
                    series = pl.from_arrow(
                        underlying.select(pl.col(root_fqn)).to_arrow()[root_fqn].cast(feature.converter.pyarrow_dtype)
                    )
                    assert isinstance(series, pl.Series)
                    underlying = underlying.with_columns(series.alias(root_fqn))
                except:
                    raise TypeError(
                        f"Expected struct field '{root_fqn}' to have dtype `{expected_dtype}`; got dtype `{actual_dtype}. Attempted to automatically coerce to the correct dtype, but failed with error: {e}`"
                    ) from e
        else:
            try:
                if actual_dtype == pl.Utf8:  # pyright: ignore[reportUnnecessaryComparison]
                    if isinstance(expected_dtype, pl.Datetime):
                        # tzinfo = None if expected_dtype.time_zone is None else zoneinfo.ZoneInfo(expected_dtype.time_zone)
                        underlying = underlying.with_columns(pl.col(root_fqn).str.strptime(pl.Datetime).alias(root_fqn))
                        if cast(pl.Datetime, underlying.schema[root_fqn]).time_zone is not None:
                            assert expected_dtype.time_zone is not None
                            cast_expr = pl.col(root_fqn).dt.convert_time_zone(expected_dtype.time_zone)
                        else:
                            cast_expr = pl.col(root_fqn).dt.replace_time_zone(expected_dtype.time_zone)
                    elif expected_dtype == pl.Date:
                        cast_expr = pl.col(root_fqn).apply(
                            lambda x: None if x is None else isodate.parse_date(x),
                        )
                    elif expected_dtype == pl.Time:
                        cast_expr = pl.col(root_fqn).apply(
                            lambda x: None if x is None else isodate.parse_time(x),
                        )
                    elif expected_dtype == pl.Duration:
                        cast_expr = pl.col(root_fqn).apply(
                            lambda x: None if x is None else isodate.parse_duration(x),
                        )
                    else:
                        cast_expr = pl.col(root_fqn).cast(expected_dtype)
                    col = cast_expr.alias(root_fqn)
                else:
                    col = pl.col(root_fqn).cast(expected_dtype)
                underlying = underlying.with_columns(col)
            except pl.ComputeError as e:
                raise TypeError(
                    (
                        f"Values for feature `{root_fqn}` with type '{feature.typ}' could not be converted "
                        f"to dtype `{expected_dtype}`. Found type {actual_dtype}, instead."
                    )
                )

    return underlying


def validate_nulls(
    underlying: Union[pl.DataFrame, pl.LazyFrame],
    missing_value_strategy: MissingValueStrategy,
    tolerate_missing_features: bool = False,
) -> pl.DataFrame:
    """Validate that any null values are in columns that support nullable values"""
    # This is called from within DataFrame.__init__, which validates that polars is installed
    import polars as pl

    if isinstance(underlying, pl.LazyFrame):
        underlying = underlying.collect()
    schema = underlying.schema
    null_count_rows = underlying.null_count().to_dicts()
    if len(null_count_rows) == 0:
        return underlying  # Empty dataframe
    null_counts = get_unique_item(null_count_rows)
    for col_name, null_count in null_counts.items():
        try:
            feature = Feature.from_root_fqn(col_name)
        except FeatureNotFoundException:
            if tolerate_missing_features:
                continue
            raise
        if null_count > 0 and not feature.typ.is_nullable and not isinstance(schema[col_name], pl.Struct):
            if missing_value_strategy == "allow":
                warnings.warn(
                    UserWarning(
                        (
                            f"Allowing missing value for feature '{col_name}' with type '{feature.typ}' strategy "
                            f"'default_or_allow'"
                        )
                    )
                )
            elif missing_value_strategy == "error":
                raise TypeError(
                    f"Feature '{col_name}' with type '{feature.typ}' has missing values, but the feature is non-nullable"
                )
            elif missing_value_strategy in ("default_or_error", "default_or_allow"):
                if feature.converter.has_default:
                    primitive_default = feature.converter.primitive_default
                    if primitive_default is not None:
                        underlying = underlying.with_columns(
                            [
                                pl.col(col_name).fill_null(
                                    value=primitive_default,
                                )
                            ]
                        )
                elif missing_value_strategy == "default_or_error":
                    raise TypeError(
                        (
                            f"Feature '{col_name}' with type '{feature.typ}' has missing values, "
                            "but the feature does not have a default"
                        )
                    )
            else:
                raise ValueError(
                    (
                        f"Unsupported missing value strategy '{missing_value_strategy}'. "
                        "Allowed options are 'allow', 'error', and 'default_or_error'."
                    )
                )
    return underlying



from typing import TYPE_CHECKING, Any, Mapping, Optional, Union

from chalk.integrations.named import create_integration_variable, load_integration_variable
from chalk.streams.base import StreamSource
from chalk.utils.duration import Duration

if TYPE_CHECKING:
    from pydantic import BaseModel
else:
    try:
        from pydantic.v1 import BaseModel
    except ImportError:
        from pydantic import BaseModel

_PUBSUB_PROJECT_ID_NAME = "PUBSUB_PROJECT_ID"
_PUBSUB_SUBSCRIPTION_ID_NAME = "PUBSUB_SUBSCRIPTION_ID"
_PUBSUB_LATE_ARRIVAL_DEADLINE_NAME = "PUBSUB_LATE_ARRIVAL_DEADLINE"
_PUBSUB_DEAD_LETTER_QUEUE_TOPIC_ID_NAME = "PUBSUB_DEAD_LETTER_QUEUE_TOPIC_ID"


class PubSubSource(StreamSource, BaseModel, frozen=True):
    project_id: Optional[str] = None
    """The project id of your PubSub source"""

    subscription_id: Optional[str] = None
    """
    The subscription id of your PubSub topic from which you want to consume messages.
    To enable permission for consuming this screen, ensure that the service account has
    the permissions 'pubsub.subscriptions.consume' and 'pubsub.subscriptions.get'.
    """

    name: Optional[str] = None
    """
    The name of the integration, as configured in your Chalk Dashboard.
    """

    late_arrival_deadline: Duration = "infinity"
    """
    Messages older than this deadline will not be processed.
    """

    dead_letter_queue_topic_id: Optional[str] = None
    """
    PubSub topic id to send messages when message processing fails. Add the permission
    'pubsub.topics.publish' if this is set.
    """

    def __init__(
        self,
        *,
        project_id: Optional[str] = None,
        subscription_id: Optional[str] = None,
        name: Optional[str] = None,
        late_arrival_deadline: Duration = "infinity",
        dead_letter_queue_topic_id: Optional[str] = None,
        integration_variable_override: Optional[Mapping[str, str]] = None,
    ):
        super(PubSubSource, self).__init__(
            name=name,
            project_id=project_id
            or load_integration_variable(
                name=_PUBSUB_PROJECT_ID_NAME, integration_name=name, override=integration_variable_override
            )
            or PubSubSource.__fields__["project_id"].default,
            subscription_id=subscription_id
            or load_integration_variable(
                name=_PUBSUB_SUBSCRIPTION_ID_NAME, integration_name=name, override=integration_variable_override
            )
            or PubSubSource.__fields__["subscription_id"].default,
            late_arrival_deadline=late_arrival_deadline
            or load_integration_variable(
                name=_PUBSUB_LATE_ARRIVAL_DEADLINE_NAME, integration_name=name, override=integration_variable_override
            )
            or PubSubSource.__fields__["late_arrival_deadline"].default,
            dead_letter_queue_topic_id=dead_letter_queue_topic_id
            or load_integration_variable(
                name=_PUBSUB_DEAD_LETTER_QUEUE_TOPIC_ID_NAME,
                integration_name=name,
                override=integration_variable_override,
            )
            or PubSubSource.__fields__["dead_letter_queue_topic_id"].default,
        )
        self.registry.append(self)

    def config_to_json(self) -> Any:
        return self.json()

    @property
    def streaming_type(self) -> str:
        return "pubsub"

    @property
    def dlq_name(self) -> Union[str, None]:
        return self.dead_letter_queue_topic_id

    @property
    def stream_or_topic_name(self) -> str:
        assert self.subscription_id is not None
        return self.subscription_id

    def _recreate_integration_variables(self) -> dict[str, str]:
        return {
            k: v
            for k, v in [
                create_integration_variable(_PUBSUB_PROJECT_ID_NAME, self.name, self.project_id),
                create_integration_variable(_PUBSUB_SUBSCRIPTION_ID_NAME, self.name, self.subscription_id),
                create_integration_variable(_PUBSUB_LATE_ARRIVAL_DEADLINE_NAME, self.name, self.late_arrival_deadline),
                create_integration_variable(
                    _PUBSUB_DEAD_LETTER_QUEUE_TOPIC_ID_NAME, self.name, self.dead_letter_queue_topic_id
                ),
            ]
            if v is not None
        }
